{$A+,B-,D+,E-,F+,G+,I-,L+,N+,O-,R-,S-,V-}

UNIT Basis;

INTERFACE

USES Maths,Extragraph;

TYPE Vector3=RECORD
        x,y,z:RealType
     END;

     Matrix3=ARRAY[1..3,1..3] OF RealType;

     RGBfloat=RECORD
        r,g,b:RealType
     END;

CONST iV:Vector3=(x:0;y:0;z:0);
      iM:Matrix3=((1,0,0),(0,1,0),(0,0,1));
      RGB0:RGBfloat=(r:0;g:0;b:0);

PROCEDURE copyRGBcol(VAR c:RGBColor;VAR cf:RGBfloat);

PROCEDURE copyRGBfloat(VAR cf:RGBfloat;VAR c:RGBColor);

PROCEDURE RGBresizeSelf(VAR cf:RGBfloat;s:RealType);
(* FPU-Optimiert *)

PROCEDURE RGBresize(VAR cf,ce:RGBfloat;s:RealType);
(* FPU-Optimiert *)

PROCEDURE RGBpartition(VAR cf1,cf2,ce:RGBfloat;s:RealType);
(* ### noch FPU-optimieren *)

PROCEDURE VnormSelf(VAR v:Vector3);
(* FPU-Optimiert *)

PROCEDURE Vnorm(VAR v,e:Vector3);
(* FPU-Optimiert *)

PROCEDURE MVmul(VAR m:Matrix3;VAR v:Vector3;VAR e:Vector3);
(* FPU-Optimiert *)

PROCEDURE MVmulSelf(VAR m:Matrix3;VAR v:Vector3);
(* FPU-Optimiert *)

PROCEDURE VVadd(VAR v1,v2:Vector3;VAR e:Vector3);
(* FPU-Optimiert *)

PROCEDURE VVsub(VAR v1,v2:Vector3;VAR e:Vector3);
(* FPU-Optimiert *)

FUNCTION Vsize(VAR v:Vector3):RealType;
(* FPU-Optimiert *)

PROCEDURE Vresize(VAR v:Vector3;s:RealType);
(* FPU-Optimiert *)

FUNCTION VVsp(VAR v1,v2:Vector3):RealType;
(* FPU-Optimiert *)

FUNCTION VVspLess0(VAR v1,v2:Vector3):BOOLEAN;
(* FPU-Optimiert *)

PROCEDURE VVcross(VAR v1,v2:Vector3;VAR e:Vector3);
(* FPU-Optimiert *)

PROCEDURE MMmul(m1,m2:Matrix3;VAR e:Matrix3);

PROCEDURE Translation(VAR m,mInverse:Matrix3;sizex,sizey,sizez,anga,angb,angc:RealType);

PROCEDURE turn2D(VAR px,py:RealType;cx,cy,ang:RealType);
(*** Dreht Punkt p um Zentrum c mit Winkel ang ***)

PROCEDURE Turn3DRight0(VAR px,py,pz:RealType;sina,cosa,sinb,cosb,sinc,cosc:RealType);
(*** Dreht 3D rechts um 0-Punkt.               ***)

PROCEDURE mirrLine(VAR n,a:Vector3;VAR b:Vector3);
(* FPU-Optimiert *)

PROCEDURE refractLine(n,a:Vector3;
                      RefractionIndex:RealType;
                      VAR b:Vector3);


IMPLEMENTATION


PROCEDURE copyRGBcol(VAR c:RGBColor;VAR cf:RGBfloat);
BEGIN
    cf.r:=c.r;
    cf.g:=c.g;
    cf.b:=c.b
END;

PROCEDURE copyRGBfloat(VAR cf:RGBfloat;VAR c:RGBColor);
BEGIN
    c.r:=ROUND(cf.r);
    c.g:=ROUND(cf.g);
    c.b:=ROUND(cf.b)
END;


(*$IFOPT N-*)

PROCEDURE RGBresizeSelf(VAR cf:RGBfloat;s:RealType);
BEGIN
    cf.r:=cf.r*s;
    cf.g:=cf.g*s;
    cf.b:=cf.b*s
END;

(*$ELSE*)

PROCEDURE RGBresizeSelf(VAR cf:RGBfloat;s:RealType);
ASSEMBLER;
ASM
                            (*  ST  |  1   |  2   |  3   |  4   |  5   |  6   |  7   *)

    FLD s                   (*  s   |      |      |      |      |      |      |      *)
    LES DI,cf
    FLD DWORD PTR ES:[DI]   (* cf.r |  s   |      |      |      |      |      |      *)
    FMUL ST,ST(1)           (* cf.r'|  s   |      |      |      |      |      |      *)
    FSTP DWORD PTR ES:[DI]  (*  s   |      |      |      |      |      |      |      *)

    FLD DWORD PTR ES:[DI+4] (* cf.g |  s   |      |      |      |      |      |      *)
    FMUL ST,ST(1)           (* cf.g'|  s   |      |      |      |      |      |      *)
    FSTP DWORD PTR ES:[DI+4](*  s   |      |      |      |      |      |      |      *)

    FLD DWORD PTR ES:[DI+8] (* cf.b |  s   |      |      |      |      |      |      *)
    FMUL                    (* cf.b'|      |      |      |      |      |      |      *)
    FSTP DWORD PTR ES:[DI+8](*      |      |      |      |      |      |      |      *)
    FWAIT
END;

(*$ENDIF*)

(*$IFOPT N+*)

PROCEDURE RGBresize(VAR cf,ce:RGBfloat;s:RealType);
BEGIN
    ce.r:=cf.r*s;
    ce.g:=cf.g*s;
    ce.b:=cf.b*s
END;

(*$ELSE*)

PROCEDURE RGBresize(VAR cf,ce:RGBfloat;s:RealType);
ASSEMBLER;
ASM
                            (*  ST  |  1   |  2   |  3   |  4   |  5   |  6   |  7   *)

    FLD s                   (*  s   |      |      |      |      |      |      |      *)
    LES DI,cf
    FLD DWORD PTR ES:[DI]   (* cf.r |  s   |      |      |      |      |      |      *)
    FMUL ST,ST(1)           (* cf.r'|  s   |      |      |      |      |      |      *)
    FLD DWORD PTR ES:[DI+4] (* cf.g | cf.r'|  s   |      |      |      |      |      *)
    FMUL ST,ST(2)           (* cf.g'| cf.r'|  s   |      |      |      |      |      *)
    FLD DWORD PTR ES:[DI+8] (* cf.b | cf.g'| cf.r'|  s   |      |      |      |      *)
    FMUL ST,ST(3)           (* cf.b'| cf.g'| cf.r'|  s   |      |      |      |      *)
    LES DI,ce
    FSTP DWORD PTR ES:[DI+8](* cf.g'| cf.r'|  s   |      |      |      |      |      *)
    FSTP DWORD PTR ES:[DI+4](* cf.r'|  s   |      |      |      |      |      |      *)
    FSTP DWORD PTR ES:[DI]  (*  s   |      |      |      |      |      |      |      *)
    FFREE ST
END;

(*$ENDIF*)



PROCEDURE RGBpartition(VAR cf1,cf2,ce:RGBfloat;s:RealType);
VAR s1:RealType;
BEGIN
    s1:=1-s;
    ce.r:=(cf1.r*s1+cf2.r*s);
    ce.g:=(cf1.g*s1+cf2.g*s);
    ce.b:=(cf1.b*s1+cf2.b*s)
END;



(*$IFOPT N-*)

PROCEDURE VnormSelf(VAR v:Vector3);
VAR l:RealType;
BEGIN
    l:=1/Sqrt(Sqr(v.x)+Sqr(v.y)+Sqr(v.z));
    v.x:=v.x*l;
    v.y:=v.y*l;
    v.z:=v.z*l
END;

(*$ELSE*)

PROCEDURE VnormSelf(VAR v:Vector3);
ASSEMBLER;
ASM
                            (*  ST  |  1   |  2   |  3   |  4   |  5   |  6   |  7   *)
    LES DI,v
    FLD DWORD PTR ES:[DI]   (*  x   |      |      |      |      |      |      |      *)
    FLD DWORD PTR ES:[DI+4] (*  y   |  x   |      |      |      |      |      |      *)
    FLD DWORD PTR ES:[DI+8] (*  z   |  y   |  x   |      |      |      |      |      *)
    FLD ST(2)               (*  x   |  z   |  y   |  x   |      |      |      |      *)
    FMUL ST,ST              (*  x^2 |  z   |  y   |  x   |      |      |      |      *)
    FLD ST(2)               (*  y   |  x^2 |  z   |  y   |  x   |      |      |      *)
    FMUL ST,ST              (*  y^2 |  x^2 |  z   |  y   |  x   |      |      |      *)
    FLD ST(2)               (*  z   |  y^2 |  x^2 |  z   |  y   |  x   |      |      *)
    FMUL ST,ST              (*  z^2 |  y^2 |  x^2 |  z   |  y   |  x   |      |      *)
    FADD
    FADD                    (*  l^2 |  z   |  y   |  x   |      |      |      |      *)
    FSQRT                   (*  l   |  z   |  y   |  x   |      |      |      |      *)
    FLD1
    FDIVR                   (*  1/l |  z   |  y   |  x   |      |      |      |      *)
    FSTP ST(4)              (*  z   |  y   |  x   |  1/l |      |      |      |      *)
    FMUL ST,ST(3)           (*  z/l |  y   |  x   |  1/l |      |      |      |      *)
    FSTP DWORD PTR ES:[DI+8](*  y   |  x   |  1/l |      |      |      |      |      *)
    FMUL ST,ST(2)           (*  y/l |  x   |  1/l |      |      |      |      |      *)
    FSTP DWORD PTR ES:[DI+4](*  x   |  1/l |      |      |      |      |      |      *)
    FMULP ST(1),ST          (*  x/l |      |      |      |      |      |      |      *)
    FSTP DWORD PTR ES:[DI]  (*      |      |      |      |      |      |      |      *)
    FWAIT
END;

(*$ENDIF*)

(*$IFOPT N-*)

PROCEDURE Vnorm(VAR v,e:Vector3);
VAR l:RealType;
BEGIN
    l:=1/Sqrt(Sqr(v.x)+Sqr(v.y)+Sqr(v.z));
    e.x:=v.x*l;
    e.y:=v.y*l;
    e.z:=v.z*l
END;

(*$ELSE*)

PROCEDURE Vnorm(VAR v,e:Vector3);
ASSEMBLER;
ASM
                            (*  ST  |  1   |  2   |  3   |  4   |  5   |  6   |  7   *)
    LES DI,v
    FLD DWORD PTR ES:[DI]   (*  x   |      |      |      |      |      |      |      *)
    FLD DWORD PTR ES:[DI+4] (*  y   |  x   |      |      |      |      |      |      *)
    FLD DWORD PTR ES:[DI+8] (*  z   |  y   |  x   |      |      |      |      |      *)
    FLD ST(2)               (*  x   |  z   |  y   |  x   |      |      |      |      *)
    FMUL ST,ST              (*  x^2 |  z   |  y   |  x   |      |      |      |      *)
    FLD ST(2)               (*  y   |  x^2 |  z   |  y   |  x   |      |      |      *)
    FMUL ST,ST              (*  y^2 |  x^2 |  z   |  y   |  x   |      |      |      *)
    FLD ST(2)               (*  z   |  y^2 |  x^2 |  z   |  y   |  x   |      |      *)
    FMUL ST,ST              (*  z^2 |  y^2 |  x^2 |  z   |  y   |  x   |      |      *)
    FADD
    FADD                    (*  l^2 |  z   |  y   |  x   |      |      |      |      *)
    FSQRT                   (*  l   |  z   |  y   |  x   |      |      |      |      *)
    FLD1
    FDIVR                   (*  1/l |  z   |  y   |  x   |      |      |      |      *)
    FSTP ST(4)              (*  z   |  y   |  x   |  1/l |      |      |      |      *)
    FMUL ST,ST(3)           (*  z/l |  y   |  x   |  1/l |      |      |      |      *)
    LES DI,e
    FSTP DWORD PTR ES:[DI+8](*  y   |  x   |  1/l |      |      |      |      |      *)
    FMUL ST,ST(2)           (*  y/l |  x   |  1/l |      |      |      |      |      *)
    FSTP DWORD PTR ES:[DI+4](*  x   |  1/l |      |      |      |      |      |      *)
    FMULP ST(1),ST          (*  x/l |      |      |      |      |      |      |      *)
    FSTP DWORD PTR ES:[DI]  (*      |      |      |      |      |      |      |      *)
    FWAIT
END;

(*$ENDIF*)


(*$IFOPT N-*)

PROCEDURE MVmul(VAR m:Matrix3;VAR v:Vector3;VAR e:Vector3);
VAR vsav:Vector3;
BEGIN
    vsav:=v;
    e.x:=vsav.x*m[1,1]+vsav.y*m[1,2]+vsav.z*m[1,3];
    e.y:=vsav.x*m[2,1]+vsav.y*m[2,2]+vsav.z*m[2,3];
    e.z:=vsav.x*m[3,1]+vsav.y*m[3,2]+vsav.z*m[3,3]
END;

(*$ELSE*)

PROCEDURE MVmul(VAR m:Matrix3;VAR v:Vector3;VAR e:Vector3);
ASSEMBLER;
ASM
                    (*  ST  |  1   |  2   |  3   |  4   |  5   |  6   |  7   *)
    LES DI,v        (*  v.x |      |      |      |      |      |      |      *)
    FLD DWORD PTR ES:[DI]
                    (*  v.y |  v.x |      |      |      |      |      |      *)
    FLD DWORD PTR ES:[DI+4]
                    (*  v.z |  v.y |  v.x |      |      |      |      |      *)
    FLD DWORD PTR ES:[DI+8]


    LES DI,m        (*  m11 | v.z  | v.y  | v.x  |      |      |      |      *)
    FLD DWORD PTR ES:[DI]
    FMUL ST,ST(3)   (*m11*vx| v.z  | v.y  | v.x  |      |      |      |      *)
                    (*  m12 |m11*vx| v.z  | v.y  | v.x  |      |      |      *)
    FLD DWORD PTR ES:[DI+4]
    FMUL ST,ST(3)   (*m12*vy|m11*vx| v.z  | v.y  | v.x  |      |      |      *)
                    (*  m13 |m12*vy|m11*vx| v.z  | v.y  | v.x  |      |      *)
    FLD DWORD PTR ES:[DI+8]
    FMUL ST,ST(3)   (*m13*vz|m12*vy|m11*vx| v.z  | v.y  | v.x  |      |      *)
    FADD
    FADD            (*  ex  | v.z  | v.y  | v.x  |      |      |      |      *)
    LES DI,e        (*  v.z | v.y  | v.x  |      |      |      |      |      *)
    FSTP DWORD PTR ES:[DI]

    LES DI,m        (*  m21 | v.z  | v.y  | v.x  |      |      |      |      *)
    FLD DWORD PTR ES:[DI+12]
    FMUL ST,ST(3)   (*m21*vx| v.z  | v.y  | v.x  |      |      |      |      *)
                    (*  m22 |m21*vx| v.z  | v.y  | v.x  |      |      |      *)
    FLD DWORD PTR ES:[DI+16]
    FMUL ST,ST(3)   (*m22*vy|m21*vx| v.z  | v.y  | v.x  |      |      |      *)
                    (*  m23 |m22*vy|m21*vx| v.z  | v.y  | v.x  |      |      *)
    FLD DWORD PTR ES:[DI+20]
    FMUL ST,ST(3)   (*m23*vz|m22*vy|m21*vx| v.z  | v.y  | v.x  |      |      *)
    FADD
    FADD            (*  ey  | v.z  | v.y  | v.x  |      |      |      |      *)
    LES DI,e        (*  v.z | v.y  | v.x  |      |      |      |      |      *)
    FSTP DWORD PTR ES:[DI+4]

    LES DI,m        (*  m31 | v.z  | v.y  | v.x  |      |      |      |      *)
    FLD DWORD PTR ES:[DI+24]
    FMUL ST,ST(3)   (*m31*vx| v.z  | v.y  | v.x  |      |      |      |      *)
                    (*  m32 |m31*vx| v.z  | v.y  | v.x  |      |      |      *)
    FLD DWORD PTR ES:[DI+28]
    FMUL ST,ST(3)   (*m32*vy|m31*vx| v.z  | v.y  | v.x  |      |      |      *)
                    (*  m33 |m32*vy|m31*vx| v.z  | v.y  | v.x  |      |      *)
    FLD DWORD PTR ES:[DI+32]
    FMUL ST,ST(3)   (*m33*vz|m32*vy|m31*vx| v.z  | v.y  | v.x  |      |      *)
    FADD
    FADD            (*  ez  | v.z  | v.y  | v.x  |      |      |      |      *)
    LES DI,e        (*  v.z | v.y  | v.x  |      |      |      |      |      *)
    FSTP DWORD PTR ES:[DI+8]
    FFREE ST
    FFREE ST(1)
    FFREE ST(2)
END;

(*$ENDIF*)

(*$IFOPT N-*)

PROCEDURE MVmulSelf(VAR m:Matrix3;VAR v:Vector3);
VAR vsav:Vector3;
BEGIN
    vsav:=v;
    v.x:=vsav.x*m[1,1]+vsav.y*m[1,2]+vsav.z*m[1,3];
    v.y:=vsav.x*m[2,1]+vsav.y*m[2,2]+vsav.z*m[2,3];
    v.z:=vsav.x*m[3,1]+vsav.y*m[3,2]+vsav.z*m[3,3]
END;

(*$ELSE*)

PROCEDURE MVmulSelf(VAR m:Matrix3;VAR v:Vector3);
ASSEMBLER;
ASM
                    (*  ST  |  1   |  2   |  3   |  4   |  5   |  6   |  7   *)
    LES DI,v        (*  v.x |      |      |      |      |      |      |      *)
    MOV AX,ES
    MOV BX,DI
    FLD DWORD PTR ES:[DI]
                    (*  v.y |  v.x |      |      |      |      |      |      *)
    FLD DWORD PTR ES:[DI+4]
                    (*  v.z |  v.y |  v.x |      |      |      |      |      *)
    FLD DWORD PTR ES:[DI+8]


    LES DI,m        (*  m11 | v.z  | v.y  | v.x  |      |      |      |      *)
    FLD DWORD PTR ES:[DI]
    FMUL ST,ST(3)   (*m11*vx| v.z  | v.y  | v.x  |      |      |      |      *)
                    (*  m12 |m11*vx| v.z  | v.y  | v.x  |      |      |      *)
    FLD DWORD PTR ES:[DI+4]
    FMUL ST,ST(3)   (*m12*vy|m11*vx| v.z  | v.y  | v.x  |      |      |      *)
                    (*  m13 |m12*vy|m11*vx| v.z  | v.y  | v.x  |      |      *)
    FLD DWORD PTR ES:[DI+8]
    FMUL ST,ST(3)   (*m13*vz|m12*vy|m11*vx| v.z  | v.y  | v.x  |      |      *)
    FADD
    FADD            (*  ex  | v.z  | v.y  | v.x  |      |      |      |      *)
    MOV ES,AX
    MOV DI,BX
    FSTP DWORD PTR ES:[DI]

    LES DI,m        (*  m21 | v.z  | v.y  | v.x  |      |      |      |      *)
    FLD DWORD PTR ES:[DI+12]
    FMUL ST,ST(3)   (*m21*vx| v.z  | v.y  | v.x  |      |      |      |      *)
                    (*  m22 |m21*vx| v.z  | v.y  | v.x  |      |      |      *)
    FLD DWORD PTR ES:[DI+16]
    FMUL ST,ST(3)   (*m22*vy|m21*vx| v.z  | v.y  | v.x  |      |      |      *)
                    (*  m23 |m22*vy|m21*vx| v.z  | v.y  | v.x  |      |      *)
    FLD DWORD PTR ES:[DI+20]
    FMUL ST,ST(3)   (*m23*vz|m22*vy|m21*vx| v.z  | v.y  | v.x  |      |      *)
    FADD
    FADD            (*  ey  | v.z  | v.y  | v.x  |      |      |      |      *)
    MOV ES,AX
    MOV DI,BX
    FSTP DWORD PTR ES:[DI+4]

    LES DI,m        (*  m31 | v.z  | v.y  | v.x  |      |      |      |      *)
    FLD DWORD PTR ES:[DI+24]
    FMUL ST,ST(3)   (*m31*vx| v.z  | v.y  | v.x  |      |      |      |      *)
                    (*  m32 |m31*vx| v.z  | v.y  | v.x  |      |      |      *)
    FLD DWORD PTR ES:[DI+28]
    FMUL ST,ST(3)   (*m32*vy|m31*vx| v.z  | v.y  | v.x  |      |      |      *)
                    (*  m33 |m32*vy|m31*vx| v.z  | v.y  | v.x  |      |      *)
    FLD DWORD PTR ES:[DI+32]
    FMUL ST,ST(3)   (*m33*vz|m32*vy|m31*vx| v.z  | v.y  | v.x  |      |      *)
    FADD
    FADD            (*  ez  | v.z  | v.y  | v.x  |      |      |      |      *)
    MOV ES,AX
    MOV DI,BX
    FSTP DWORD PTR ES:[DI+8]
    FFREE ST
    FFREE ST(1)
    FFREE ST(2)
END;

(*$ENDIF*)


(*$IFOPT N-*)

PROCEDURE VVadd(VAR v1,v2:Vector3;VAR e:Vector3);
BEGIN
    e.x:=v1.x+v2.x;
    e.y:=v1.y+v2.y;
    e.z:=v1.z+v2.z
END;

(*$ELSE*)

PROCEDURE VVadd(VAR v1,v2:Vector3;VAR e:Vector3);
ASSEMBLER;
ASM
                            (*  ST  |  1   |  2   |  3   |  4   |  5   |  6   |  7   *)
    LES DI,v1
    FLD DWORD PTR ES:[DI]   (* v1.x |      |      |      |      |      |      |      *)
    FLD DWORD PTR ES:[DI+4] (* v1.y | v1.x |      |      |      |      |      |      *)
    FLD DWORD PTR ES:[DI+8] (* v1.z | v1.y | v1.x |      |      |      |      |      *)

    LES DI,v2               
    FLD DWORD PTR ES:[DI]   (* v2.x | v1.z | v1.y | v1.x |      |      |      |      *)
    FLD DWORD PTR ES:[DI+4] (* v2.y | v2.x | v1.z | v1.y | v1.x |      |      |      *)
    FLD DWORD PTR ES:[DI+8] (* v2.z | v2.y | v2.x | v1.z | v1.y | v1.x |      |      *)
    LES DI,e

    FADD ST,ST(3)           (* e.z  | v2.y | v2.x | v1.z | v1.y | v1.x |      |      *)
    FSTP DWORD PTR ES:[DI+8](* v2.y | v2.x | v1.z | v1.y | v1.x |      |      |      *)
    FADD ST,ST(3)           (* e.y  | v2.x | v1.z | v1.y | v1.x |      |      |      *)
    FSTP DWORD PTR ES:[DI+4](* v2.x | v1.z | v1.y | v1.x |      |      |      |      *)
    FADD ST,ST(3)           (* e.x  | v1.z | v1.y | v1.x |      |      |      |      *)
    FSTP DWORD PTR ES:[DI]  (* v1.z | v1.y | v1.x |      |      |      |      |      *)
    FFREE ST
    FFREE ST(1)
    FFREE ST(2)
END;

(*$ENDIF*)


(*$IFOPT N-*)


PROCEDURE VVsub(VAR v1,v2:Vector3;VAR e:Vector3);
BEGIN
    e.x:=v1.x-v2.x;
    e.y:=v1.y-v2.y;
    e.z:=v1.z-v2.z
END;

(*$ELSE*)

PROCEDURE VVsub(VAR v1,v2:Vector3;VAR e:Vector3);
ASSEMBLER;
ASM
                            (*  ST  |  1   |  2   |  3   |  4   |  5   |  6   |  7   *)
    LES DI,v1
    FLD DWORD PTR ES:[DI]   (* v1.x |      |      |      |      |      |      |      *)
    FLD DWORD PTR ES:[DI+4] (* v1.y | v1.x |      |      |      |      |      |      *)
    FLD DWORD PTR ES:[DI+8] (* v1.z | v1.y | v1.x |      |      |      |      |      *)

    LES DI,v2               
    FLD DWORD PTR ES:[DI]   (* v2.x | v1.z | v1.y | v1.x |      |      |      |      *)
    FLD DWORD PTR ES:[DI+4] (* v2.y | v2.x | v1.z | v1.y | v1.x |      |      |      *)
    FLD DWORD PTR ES:[DI+8] (* v2.z | v2.y | v2.x | v1.z | v1.y | v1.x |      |      *)
    LES DI,e

    FSUBR ST,ST(3)          (* e.z  | v2.y | v2.x | v1.z | v1.y | v1.x |      |      *)
    FSTP DWORD PTR ES:[DI+8](* v2.y | v2.x | v1.z | v1.y | v1.x |      |      |      *)
    FSUBR ST,ST(3)          (* e.y  | v2.x | v1.z | v1.y | v1.x |      |      |      *)
    FSTP DWORD PTR ES:[DI+4](* v2.x | v1.z | v1.y | v1.x |      |      |      |      *)
    FSUBR ST,ST(3)          (* e.x  | v1.z | v1.y | v1.x |      |      |      |      *)
    FSTP DWORD PTR ES:[DI]  (* v1.z | v1.y | v1.x |      |      |      |      |      *)
    FFREE ST
    FFREE ST(1)
    FFREE ST(2)
END;

(*$ENDIF*)

(*$IFOPT N-*)

FUNCTION Vsize(VAR v:Vector3):RealType;
BEGIN
    Vsize:=Sqrt(Sqr(v.x)+Sqr(v.y)+Sqr(v.z))
END;

(*$ELSE*)

FUNCTION Vsize(VAR v:Vector3):RealType;
ASSEMBLER;
ASM
                            (*  ST  |  1   |  2   |  3   |  4   |  5   |  6   |  7   *)
    LES DI,v
    FLD DWORD PTR ES:[DI]   (* v.x  |      |      |      |      |      |      |      *)
    FMUL ST,ST              (* v.x^2|      |      |      |      |      |      |      *)
    FLD DWORD PTR ES:[DI+4] (* v.y  | v.x^2|      |      |      |      |      |      *)
    FMUL ST,ST              (* v.y^2| v.x^2|      |      |      |      |      |      *)
    FLD DWORD PTR ES:[DI+8] (* v.z  | v.y^2| v.x^2|      |      |      |      |      *)
    FMUL ST,ST              (* v.z^2| v.y^2| v.x^2|      |      |      |      |      *)
    FADD
    FADD                    (*size^2|      |      |      |      |      |      |      *)
    FSQRT                   (* size |      |      |      |      |      |      |      *)
    FWAIT
END;

(*$ENDIF*)

(*$IFOPT N-*)

PROCEDURE Vresize(VAR v:Vector3;s:RealType);
BEGIN
    v.x:=v.x*s;
    v.y:=v.y*s;
    v.z:=v.z*s
END;

(*$ELSE*)

PROCEDURE Vresize(VAR v:Vector3;s:RealType);
ASSEMBLER;
ASM
                            (*  ST  |  1   |  2   |  3   |  4   |  5   |  6   |  7   *)

    FLD s                   (*  s   |      |      |      |      |      |      |      *)
    LES DI,v
    FLD DWORD PTR ES:[DI]   (* v.x  |  s   |      |      |      |      |      |      *)
    FMUL ST,ST(1)           (* v.x' |  s   |      |      |      |      |      |      *)
    FSTP DWORD PTR ES:[DI]  (*  s   |      |      |      |      |      |      |      *)

    FLD DWORD PTR ES:[DI+4] (* v.y  |  s   |      |      |      |      |      |      *)
    FMUL ST,ST(1)           (* v.y' |  s   |      |      |      |      |      |      *)
    FSTP DWORD PTR ES:[DI+4](*  s   |      |      |      |      |      |      |      *)

    FLD DWORD PTR ES:[DI+8] (* v.y  |  s   |      |      |      |      |      |      *)
    FMUL                    (* v.y' |      |      |      |      |      |      |      *)
    FSTP DWORD PTR ES:[DI+8](*      |      |      |      |      |      |      |      *)
    FWAIT
END;

(*$ENDIF*)

(*$IFOPT N-*)

FUNCTION VVsp(VAR v1,v2:Vector3):RealType;
BEGIN
    VVsp:=v1.x*v2.x+v1.y*v2.y+v1.z*v2.z
END;

(*$ELSE*)

FUNCTION VVsp(VAR v1,v2:Vector3):RealType;
ASSEMBLER;
ASM
                            (*  ST  |  1   |  2   |  3   |  4   |  5   |  6   |  7   *)
    LES DI,v1
    FLD DWORD PTR ES:[DI]   (* v1.x |      |      |      |      |      |      |      *)
    FLD DWORD PTR ES:[DI+4] (* v1.y | v1.x |      |      |      |      |      |      *)
    FLD DWORD PTR ES:[DI+8] (* v1.z | v1.y | v1.x |      |      |      |      |      *)

    LES DI,v2
    FLD DWORD PTR ES:[DI]   (* v2.x | v1.z | v1.y | v1.x |      |      |      |      *)
    FMUL ST,ST(3)           (*v2*1.x| v1.z | v1.y | v1.x |      |      |      |      *)

    FLD DWORD PTR ES:[DI+4] (* v2.y |v2*1.x| v1.z | v1.y | v1.x |      |      |      *)
    FMUL ST,ST(3)           (*v2*1.y|v2*1.x| v1.z | v1.y | v1.x |      |      |      *)

    FLD DWORD PTR ES:[DI+8] (* v2.z | v2.y |v2*1.x| v1.z | v1.y | v1.x |      |      |      *)
    FMUL ST,ST(3)           (*v2*1.z|v2*1.y|v2*1.x| v1.z | v1.y | v1.x |      |      |      *)

    FADD
    FADD
    FFREE ST(1)
    FFREE ST(2)
    FFREE ST(3)
END;

(*$ENDIF*)

(*$IFOPT N-*)

FUNCTION VVspLess0(VAR v1,v2:Vector3):BOOLEAN;
BEGIN
    VVspLess0:=((v1.x*v2.x+v1.y*v2.y+v1.z*v2.z)<0)
END;

(*$ELSE*)

FUNCTION VVspLess0(VAR v1,v2:Vector3):BOOLEAN;
ASSEMBLER;
ASM
                            (*  ST  |  1   |  2   |  3   |  4   |  5   |  6   |  7   *)
    LES DI,v1
    FLD DWORD PTR ES:[DI]   (* v1.x |      |      |      |      |      |      |      *)
    FLD DWORD PTR ES:[DI+4] (* v1.y | v1.x |      |      |      |      |      |      *)
    FLD DWORD PTR ES:[DI+8] (* v1.z | v1.y | v1.x |      |      |      |      |      *)

    LES DI,v2
    FLD DWORD PTR ES:[DI]   (* v2.x | v1.z | v1.y | v1.x |      |      |      |      *)
    FMUL ST,ST(3)           (*v2*1.x| v1.z | v1.y | v1.x |      |      |      |      *)

    FLD DWORD PTR ES:[DI+4] (* v2.y |v2*1.x| v1.z | v1.y | v1.x |      |      |      *)
    FMUL ST,ST(3)           (*v2*1.y|v2*1.x| v1.z | v1.y | v1.x |      |      |      *)

    FLD DWORD PTR ES:[DI+8] (* v2.z | v2.y |v2*1.x| v1.z | v1.y | v1.x |      |      |      *)
    FMUL ST,ST(3)           (*v2*1.z|v2*1.y|v2*1.x| v1.z | v1.y | v1.x |      |      |      *)

    FADD
    FADD
    FTST
    FSTSW WORD PTR @@3
    FWAIT
    MOV AX,WORD PTR @@3
    SAHF
    JAE @@1
    MOV AL,1
    JMP @@2
@@3:DW  0
@@1:MOV AL,0
@@2:FFREE ST
    FFREE ST(1)
    FFREE ST(2)
    FFREE ST(3)
END;

(*$ENDIF*)

(*$IFOPT N-*)

PROCEDURE VVcross(VAR v1,v2:Vector3;VAR e:Vector3);
BEGIN
     e.x:=v1.y*v2.z-v1.z*v2.y;
     e.y:=v1.z*v2.x-v1.x*v2.z;
     e.z:=v1.x*v2.y-v1.y*v2.x
END;


(*$ELSE*)

PROCEDURE VVcross(VAR v1,v2:Vector3;VAR e:Vector3);
ASSEMBLER;
ASM
                             (*  ST  |  1   |  2   |  3   |  4   |  5   |  6   |  7   *)
    LES DI,v1                
    FLD DWORD PTR ES:[DI]    (* v1x  |      |      |      |      |      |      |      *)
    FLD DWORD PTR ES:[DI+4]  (* v1y  | v1x  |      |      |      |      |      |      *)
    FLD DWORD PTR ES:[DI+8]  (* v1z  | v1y  | v1x  |      |      |      |      |      *)
    LES DI,v2
    FLD DWORD PTR ES:[DI]    (* v2x  | v1z  | v1y  | v1x  |      |      |      |      *)
    FLD DWORD PTR ES:[DI+4]  (* v2y  | v2x  | v1z  | v1y  | v1x  |      |      |      *)
    FLD DWORD PTR ES:[DI+8]  (* v2z  | v2y  | v2x  | v1z  | v1y  | v1x  |      |      *)

    FLD ST(4)                (* v1y  | v2z  | v2y  | v2x  | v1z  | v1y  | v1x  |      *)
    FMUL ST,ST(1)            (*v1yv2z| v2z  | v2y  | v2x  | v1z  | v1y  | v1x  |      *)
    FLD ST(4)                (* v1z  |v1yv2z| v2z  | v2y  | v2x  | v1z  | v1y  | v1x  *)
    FMUL ST,ST(3)            (*v1zv2y|v1yv2z| v2z  | v2y  | v2x  | v1z  | v1y  | v1x  *)
    FSUB                     (* e.x  | v2z  | v2y  | v2x  | v1z  | v1y  | v1x  |      *)
    LES DI,e
    FSTP DWORD PTR ES:[DI]   (* v2z  | v2y  | v2x  | v1z  | v1y  | v1x  |      |      *)

    FLD ST(3)                (* v1z  | v2z  | v2y  | v2x  | v1z  | v1y  | v1x  |      *)
    FMUL ST,ST(3)            (*v1zv2x| v2z  | v2y  | v2x  | v1z  | v1y  | v1x  |      *)
    FLD ST(6)                (* v1x  |v1yv2z| v2z  | v2y  | v2x  | v1z  | v1y  | v1x  *)
    FMUL ST,ST(2)            (*v1xv2z|v1yv2z| v2z  | v2y  | v2x  | v1z  | v1y  | v1x  *)
    FSUB                     (* e.y  | v2z  | v2y  | v2x  | v1z  | v1y  | v1x  |      *)
    FSTP DWORD PTR ES:[DI+4] (* v2z  | v2y  | v2x  | v1z  | v1y  | v1x  |      |      *)

    FLD ST(5)                (* v1x  | v2z  | v2y  | v2x  | v1z  | v1y  | v1x  |      *)
    FMUL ST,ST(2)            (*v1xv2y| v2z  | v2y  | v2x  | v1z  | v1y  | v1x  |      *)
    FLD ST(5)                (* v1y  |v1yv2z| v2z  | v2y  | v2x  | v1z  | v1y  | v1x  *)
    FMUL ST,ST(4)            (*v1yv2x|v1yv2z| v2z  | v2y  | v2x  | v1z  | v1y  | v1x  *)
    FSUB                     (* e.z  | v2z  | v2y  | v2x  | v1z  | v1y  | v1x  |      *)
    FSTP DWORD PTR ES:[DI+8] (* v2z  | v2y  | v2x  | v1z  | v1y  | v1x  |      |      *)
    FFREE ST
    FFREE ST(1)
    FFREE ST(2)
    FFREE ST(3)
    FFREE ST(4)
    FFREE ST(5)

END;

(*$ENDIF*)

(*$IFOPT N-*)

PROCEDURE Turn3DRight0(VAR px,py,pz:RealType;sina,cosa,sinb,cosb,sinc,cosc:RealType);
VAR dx2,dy2:RealType;
BEGIN
    dx2:=px*cosc+py*sinc;
    dy2:=-px*sinc+py*cosc;

    py:=dy2*cosb+pz*sinb;
    pz:=-dy2*sinb+pz*cosb;

    px:=dx2*cosa+pz*sina;
    pz:=-dx2*sina+pz*cosa
END;

(*$ELSE*)

PROCEDURE Turn3DRight0(VAR px,py,pz:RealType;sina,cosa,sinb,cosb,sinc,cosc:RealType);
ASSEMBLER;
ASM
                    (*  ST  |  1   |  2   |  3   |  4   |  5   |  6   |  7   *)
    LES DI,px       (*  px  |      |      |      |      |      |      |      *)
    FLD DWORD PTR ES:[DI]
    FLD cosc        (* cosc |  px  |      |      |      |      |      |      *)
    FST ST(2)       (* cosc |  px  | cosc |      |      |      |      |      *)
    FLD sinc        (* sinc | cosc |  px  | cosc |      |      |      |      *)
    FST ST(4)       (* sinc | cosc |  px  | cosc | sinc |      |      |      *)
    FMUL ST,ST(2)   (*sincpx| cosc |  px  | cosc | sinc |      |      |      *)
    FXCH ST(2)      (*  px  | cosc |sincpx| cosc | sinc |      |      |      *)
    FMUL            (*coscpx|sincpx| cosc | sinc |      |      |      |      *)
    FSTP ST(4)      (*sincpx| cosc | sinc |coscpx|      |      |      |      *)
    FSTP ST(4)      (* cosc | sinc |coscpx|sincpx|      |      |      |      *)
    LES DI,py       (*  py  | cosc | sinc |coscpx|sincpx|      |      |      *)
    FLD DWORD PTR ES:[DI]
    FMUL ST(2),ST   (*  py  | cosc |sincpy|coscpx|sincpx|      |      |      *)
    FMUL            (*coscpy|sincpy|coscpx|sincpx|      |      |      |      *)
    FXCH ST(2)      (*coscpx|sincpy|coscpy|sincpx|      |      |      |      *)
    FADD            (*  x2  |coscpx|sincpx|      |      |      |      |      *)
    FSTP ST(3)      (*coscpx|sincpx|  x2  |      |      |      |      |      *)
    FSUBR           (*  y2  |  x2  |      |      |      |      |      |      *)

                    (*  ST  |  1   |  2   |  3   |  4   |  5   |  6   |  7   *)
    FLD cosb        (* cosb |  y2  |  x2  |      |      |      |      |      *)
    FST ST(3)       (* cosb |  y2  |  x2  | cosb |      |      |      |      *)
    FLD sinb        (* sinb | cosb |  y2  |  x2  | cosb |      |      |      *)
    FST ST(5)       (* sinb | cosb |  y2  |  x2  | cosb | sinb |      |      *)
    FMUL ST,ST(2)   (*sinby2| cosb |  y2  |  x2  | cosb | sinb |      |      *)
    FXCH ST(2)      (*  y2  | cosb |sinby2|  x2  | cosb | sinb |      |      *)
    FMUL            (*cosby2|sinby2|  x2  | cosb | sinb |      |      |      *)
    FSTP ST(5)      (*sinby2|  x2  | cosb | sinb |cosby2|      |      |      *)
    FSTP ST(5)      (*  x2  | cosb | sinb |cosby2|sinby2|      |      |      *)
    FSTP ST(5)      (* cosb | sinb |cosby2|sinby2|  x2  |      |      |      *)
    LES DI,pz       (*  pz  | cosb | sinb |cosby2|sinby2|  x2  |      |      *)
    FLD DWORD PTR ES:[DI]
    FMUL ST(2),ST   (*  pz  | cosb |sinbpz|cosby2|sinby2|  x2  |      |      *)
    FMUL            (*cosbpz|sinbpz|cosby2|sinby2|  x2  |      |      |      *)
    FXCH ST(2)      (*cosby2|sinbpz|cosbpz|sinby2|  x2  |      |      |      *)
    FADD            (*  y3  |cosbpz|sinby2|  x2  |      |      |      |      *)
    LES DI,py       (*cosbpz|sinby2|  x2  |      |      |      |      |      *)
    FSTP DWORD PTR ES:[DI]
    FSUBR           (*  z2  |  x2  |      |      |      |      |      |      *)
    FSTP ST(2)      (*  x2  |  z2  |      |      |      |      |      |      *)

                    (*  ST  |  1   |  2   |  3   |  4   |  5   |  6   |  7   *)
    FLD cosa        (* cosa |  x2  |  z2  |      |      |      |      |      *)
    FST ST(3)       (* cosa |  x2  |  z2  | cosa |      |      |      |      *)
    FLD sina        (* sina | cosa |  x2  |  z2  | cosa |      |      |      *)
    FST ST(5)       (* sina | cosa |  x2  |  z2  | cosa | sina |      |      *)
    FMUL ST,ST(2)   (*sinax2| cosa |  x2  |  z2  | cosa | sina |      |      *)
    FXCH ST(2)      (*  x2  | cosa |sinax2|  z2  | cosa | sina |      |      *)
    FMUL            (*cosax2|sinax2|  z2  | cosa | sina |      |      |      *)
    FSTP ST(5)      (*sinax2|  z2  | cosa | sina |cosax2|      |      |      *)
    FSTP ST(5)      (*  z2  | cosa | sina |cosax2|sinax2|      |      |      *)
    FMUL ST(2),ST   (*  z2  | cosa |sinaz2|cosax2|sinax2|      |      |      *)
    FMUL            (*cosaz2|sinaz2|cosax2|sinax2|      |      |      |      *)
    FXCH ST(2)      (*cosax2|sinaz2|cosaz2|sinax2|      |      |      |      *)
    FADD            (*  x3  |cosaz2|sinax2|      |      |      |      |      *)
    LES DI,px       (*cosaz2|sinax2|      |      |      |      |      |      *)
    FSTP DWORD PTR ES:[DI]
    FSUBR           (*  z2  |      |      |      |      |      |      |      *)
    LES DI,pz       (*      |      |      |      |      |      |      |      *)
    FSTP DWORD PTR ES:[DI]
    WAIT
END;

(*$ENDIF*)


PROCEDURE MMmul(m1,m2:Matrix3;VAR e:Matrix3);
VAR i,j:BYTE;
BEGIN
    FOR i:=1 TO 3 DO
        FOR j:=1 TO 3 DO
           e[i,j]:=m1[i,1]*m2[1,j]+m1[i,2]*m2[2,j]+m1[i,3]*m2[3,j]
END;


PROCEDURE Translation(VAR m,mInverse:Matrix3;sizex,sizey,sizez,anga,angb,angc:RealType);
VAR i,j:WORD;
    ms,msI:Matrix3;
    ma,maI:Matrix3;
    mb,mbI:Matrix3;
    mc,mcI:Matrix3;
    sina,cosa,sinb,cosb,sinc,cosc:RealType;
BEGIN
    FOR i:=1 TO 3 DO
        FOR j:=1 TO 3 DO
            IF i=j THEN
                CASE i OF
                    1:ms[i,j]:=1/sizex;
                    2:ms[i,j]:=1/sizey;
                    3:ms[i,j]:=1/sizez
                END
            ELSE
                ms[i,j]:=0;
    msI:=ms;

    anga:=anga*PiDiv180;
    angb:=angb*PiDiv180;
    angc:=angc*PiDiv180;
    sina:=Sin(anga);
    cosa:=Cos(anga);
    sinb:=Sin(angb);
    cosb:=Cos(angb);
    sinc:=Sin(angc);
    cosc:=Cos(angc);

    ma [1,1]:= cosa;    ma [1,2]:=    0;    ma [1,3]:=-sina;
    ma [2,1]:=    0;    ma [2,2]:=    1;    ma [2,3]:=    0;
    ma [3,1]:= sina;    ma [3,2]:=    0;    ma [3,3]:= cosa;

    mb [1,1]:=    1;    mb [1,2]:=    0;    mb [1,3]:=    0;
    mb [2,1]:=    0;    mb [2,2]:= cosb;    mb [2,3]:=-sinb;
    mb [3,1]:=    0;    mb [3,2]:= sinb;    mb [3,3]:= cosb;

    mc [1,1]:= cosc;    mc [1,2]:=-sinc;    mc [1,3]:=    0;
    mc [2,1]:= sinc;    mc [2,2]:= cosc;    mc [2,3]:=    0;
    mc [3,1]:=    0;    mc [3,2]:=    0;    mc [3,3]:=    1;

    maI[1,1]:= cosa;    maI[1,2]:=    0;    maI[1,3]:= sina;
    maI[2,1]:=    0;    maI[2,2]:=    1;    maI[2,3]:=    0;
    maI[3,1]:=-sina;    maI[3,2]:=    0;    maI[3,3]:= cosa;

    mbI[1,1]:=    1;    mbI[1,2]:=    0;    mbI[1,3]:=    0;
    mbI[2,1]:=    0;    mbI[2,2]:= cosb;    mbI[2,3]:= sinb;
    mbI[3,1]:=    0;    mbI[3,2]:=-sinb;    mbI[3,3]:= cosb;

    mcI[1,1]:= cosc;    mcI[1,2]:= sinc;    mcI[1,3]:=    0;
    mcI[2,1]:=-sinc;    mcI[2,2]:= cosc;    mcI[2,3]:=    0;
    mcI[3,1]:=    0;    mcI[3,2]:=    0;    mcI[3,3]:=    1;

    MMmul(ms,mc,m);
    MMmul(m,mb,m);
    MMmul(m,ma,m);

    MMmul(maI,mbI,mInverse);
    MMmul(mInverse,mcI,mInverse);
    MMmul(mInverse,msI,mInverse)

END;

PROCEDURE turn2D(VAR px,py:RealType;cx,cy,ang:RealType);
VAR dx,dy,dx2,dy2:RealType;
    sina,cosa:RealType;
BEGIN
    dx:=px-cx;
    dy:=py-cy;
    sina:=Sin(ang);
    cosa:=Cos(ang);
    dx2:=dx*cosa+dy*sina;
    dy2:=-dx*sina+dy*cosa;
    px:=cx+dx2;
    py:=cy+dy2
END;


(*$IFOPT N-*)

PROCEDURE mirrLine(VAR n,a:Vector3;VAR b:Vector3);
VAR d:RealType;
    dd:Vector3;
BEGIN
    d:=2/(Sqr(n.x)+Sqr(n.y)+Sqr(n.z));

    VVcross(a,n,dd);
    VVcross(n,dd,b);

    b.x:=b.x*d-a.x;
    b.y:=b.y*d-a.y;
    b.z:=b.z*d-a.z
END;

(*$ELSE*)

PROCEDURE mirrLine(VAR n,a:Vector3;VAR b:Vector3);
VAR d:RealType;
    dd:Vector3;
BEGIN
    d:=2/(Sqr(n.x)+Sqr(n.y)+Sqr(n.z));

    VVcross(a,n,dd);
    VVcross(n,dd,b);

    ASM
                             (*  ST  |  1   |  2   |  3   |  4   |  5   |  6   |  7   *)
    LES DI,a                 
    FLD DWORD PTR ES:[DI]    (* a.x  |      |      |      |      |      |      |      *)
    FLD DWORD PTR ES:[DI+4]  (* a.y  | a.x  |      |      |      |      |      |      *)
    FLD DWORD PTR ES:[DI+8]  (* a.z  | a.y  | a.x  |      |      |      |      |      *)
    FLD d                    (*  d   | a.z  | a.y  | a.x  |      |      |      |      *)
    LES DI,b

    FLD DWORD PTR ES:[DI]    (* b.x  |  d   | a.z  | a.y  | a.x  |      |      |      *)
    FMUL ST,ST(1)            (* d*b.x|  d   | a.z  | a.y  | a.x  |      |      |      *)
    FSUB ST,ST(4)            (* b.x' |  d   | a.z  | a.y  | a.x  |      |      |      *)
    FSTP DWORD PTR ES:[DI]   (*  d   | a.z  | a.y  | a.x  |      |      |      |      *)

    FLD DWORD PTR ES:[DI+4]  (* b.y  |  d   | a.z  | a.y  | a.x  |      |      |      *)
    FMUL ST,ST(1)            (* d*b.y|  d   | a.z  | a.y  | a.x  |      |      |      *)
    FSUB ST,ST(3)            (* b.y' |  d   | a.z  | a.y  | a.x  |      |      |      *)
    FSTP DWORD PTR ES:[DI+4] (*  d   | a.z  | a.y  | a.x  |      |      |      |      *)

    FLD DWORD PTR ES:[DI+8]  (* b.z  |  d   | a.z  | a.y  | a.x  |      |      |      *)
    FMUL ST,ST(1)            (* d*b.z|  d   | a.z  | a.y  | a.x  |      |      |      *)
    FSUB ST,ST(2)            (* b.z' |  d   | a.z  | a.y  | a.x  |      |      |      *)
    FSTP DWORD PTR ES:[DI+8] (*  d   | a.z  | a.y  | a.x  |      |      |      |      *)
    FFREE ST
    FFREE ST(1)
    FFREE ST(2)
    FFREE ST(3)
    END;

END;

(*$ENDIF*)

PROCEDURE refractLine(n,a:Vector3;
                      RefractionIndex:RealType;
                      VAR b:Vector3);

VAR d1,d2,d3,d:RealType;
BEGIN
    d:=(n.x*a.x+n.y*a.y+n.z*a.z);
    IF ABS(d)<0.0001 THEN
    BEGIN
        b.x:=a.x;
        b.y:=a.y;
        b.z:=a.z;
        Exit
    END;
    d:=-(Sqr(a.x)+Sqr(a.y)+Sqr(a.z))/d;
    d1:=(a.x+n.x*d)/RefractionIndex;
    d2:=(a.y+n.y*d)/RefractionIndex;
    d3:=(a.z+n.z*d)/RefractionIndex;
    b.x:=d1-n.x*d;
    b.y:=d2-n.y*d;
    b.z:=d3-n.z*d
END;

END.
