{$A+,B-,D+,E-,F+,G+,I-,L+,N+,O-,R-,S-,V-}
{$DEFINE PROFILERx}
{$DEFINE FigNamesx}

UNIT Figures;

INTERFACE

USES Maths,Basis,Messages,ObjPoints,Intervalls,SolidObjects
{$IFDEF PROFILER}
     ,Profiler
{$ENDIF}
     ;


CONST maxPartFig=10;

TYPE OperationType=(
     (*** Binre Operationstypen fr die Klasse PBuildFig ***)
       OrOp,           (* A + B                         *)
       AndOp,          (* A * B                         *)
       AndNotOp,       (* A - B                         *)
       OrOpSmooth,     (* A + B mit abgerundeten Kanten *)
       AndOpSmooth,    (* A * B mit abgerundeten Kanten *)
       AndNotOpSmooth  (* A - B mit abgerundeten Kanten *)
     );

     (*** Die Operationen ..Smooth werden gem r=3..n  ***)
     (*** berechnet: je grer desto weniger gerundet   ***)

     FigKind=(
         Solid,
         Build
     );



     PartFig=RECORD
         kind:FigKind;
         fig:POINTER
     END;

     PBuildFigur=^BuildFigur;
     BuildFigur=OBJECT
         c:Vector3;
         m:Matrix3;
         mInverse:Matrix3;

         box:BOOLEAN;
         boxFromX,boxToX,boxFromY,boxToY,boxFromZ,boxToZ:RealType;

         rSmooth:RealType;

         operation:OperationType;

         figCnt:WORD;

         figs:ARRAY[1..maxPartFig] OF PartFig;

{$IFDEF FigNames}
         name:STRING20;
{$ENDIF}
         PROCEDURE GetIntervall(VAR p,r:Vector3;VAR resultIntervall:IntervallSet);
     END;


IMPLEMENTATION



PROCEDURE BuildFigur.GetIntervall(VAR p,r:Vector3;VAR resultIntervall:IntervallSet);
VAR int1,int2:IntervallSet;
    p2,r2:Vector3;
    i:WORD;
    continue:BOOLEAN;
    nn,nn2:RealType;

BEGIN
{$IFDEF PROFILER}
    BeginIt(6);
{$ENDIF}


    IF box THEN
    BEGIN
        IF r.x>0 THEN
            nn:=(boxFromX-p.x)/(r.x+0.1)
        ELSE
            nn:=(boxToX-p.x)/(r.x-0.1);

        IF r.y>0 THEN
            nn2:=(boxFromY-p.y)/(r.y+0.1)
        ELSE
            nn2:=(boxToY-p.y)/(r.y-0.1);
        IF nn2>nn THEN
            nn:=nn2;

        IF r.z>0 THEN
            nn2:=(boxFromZ-p.z)/(r.z+0.1)
        ELSE
            nn2:=(boxToZ-p.z)/(r.z-0.1);
        IF nn2>nn THEN
            nn:=nn2;

        IF r.x>0 THEN
        BEGIN
            IF p.x+nn*r.x>boxToX THEN
            BEGIN
                resultIntervall.cnt:=0;
                Exit
            END
        END
        ELSE BEGIN
            IF p.x+nn*r.x<boxFromX THEN
            BEGIN
                resultIntervall.cnt:=0;
                Exit
            END
        END;

        IF r.y>0 THEN
        BEGIN
            IF p.y+nn*r.y>boxToY THEN
            BEGIN
                resultIntervall.cnt:=0;
                Exit
            END
        END
        ELSE BEGIN
            IF p.y+nn*r.y<boxFromY THEN
            BEGIN
                resultIntervall.cnt:=0;
                Exit
            END
        END;

        IF r.z>0 THEN
        BEGIN
            IF p.z+nn*r.z>boxToZ THEN
            BEGIN
                resultIntervall.cnt:=0;
                Exit
            END
        END
        ELSE BEGIN
            IF p.z+nn*r.z<boxFromZ THEN
            BEGIN
                resultIntervall.cnt:=0;
                Exit
            END
        END
    END;



    VVsub(p,c,p2);
    MVmulSelf(m,p2);
    MVmul(m,r,r2);

    WITH figs[1] DO
    BEGIN
        IF kind=Solid THEN
            PSolidObject(fig)^.GetIntervall(p2,r2,resultIntervall)
        ELSE
            PBuildFigur(fig)^.GetIntervall(p2,r2,resultIntervall)
    END;
    continue:=TRUE;
    i:=2;
    WHILE continue AND (i<=figCnt) DO
    BEGIN
        WITH figs[i] DO
        BEGIN
            IF kind=Solid THEN
                PSolidObject(fig)^.GetIntervall(p2,r2,int2)
            ELSE
                PBuildFigur(fig)^.GetIntervall(p2,r2,int2)
        END;
        int1:=resultIntervall;

        CASE operation OF
            OrOp:BEGIN
                     WITH resultIntervall,block[1] DO
                         IF (cnt=1) AND
                            (iFrom.l<=-oInfinite) AND
                            (iTo.l>=oInfinite) THEN
                             continue:=FALSE
                         ELSE
                             resultIntervall.OrOp(int1,int2)
                 END;
            AndOp:
                 BEGIN
                     IF resultIntervall.cnt=0 THEN
                         continue:=FALSE
                     ELSE
                         resultIntervall.AndOp(int1,int2);
                 END;
            AndNotOp:
                 BEGIN
                     IF resultIntervall.cnt=0 THEN
                         continue:=FALSE
                     ELSE
                         resultIntervall.AndNotOp(int1,int2)
                 END;
            OrOpSmooth:
                 resultIntervall.OrOpSmooth(int1,int2,rSmooth);
            AndOpSmooth:
                 resultIntervall.AndOpSmooth(int1,int2,rSmooth);
            AndNotOpSmooth:
                 resultIntervall.AndNotOpSmooth(int1,int2,rSmooth);
        END;
        INC(i)
        
    END;
    resultIntervall.ReNormal(mInverse)
{$IFDEF PROFILER}
    ;EndIt(6)
{$ENDIF}
END;


END.
